/*-
 * Copyright (c) 2010 Lev Serebryakov <lev@FreeBSD.org>
 * Copyright (c) 2006 Arne Wrner <arne_woerner@yahoo.com>
 * testing + tuning-tricks: veronica@fluffles.net
 * derived from gstripe/gmirror (Pawel Jakub Dawidek <pjd@FreeBSD.org>)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Historic Id: g_raid5.h,v 1.12.1.34.1.9.1.3 2007/08/20 20:13:58 aw Exp aw
 *
 * $Id: g_raid5.h,v 1b6df7194146 2011/08/31 19:06:41 lev $
 */

#ifndef	_G_RAID5_H_
#define	_G_RAID5_H_

#include <sys/endian.h>

#define	G_RAID5_CLASS_NAME	"RAID5"
#define	G_RAID5_MAGIC		"GEOM::RAID5"

/*
 * Version history:
 * 1 - Initial version number.
 * 2 - quite stable + provider size checking
 * 3 - experimental + fast combining but high kmem use
 */
#define	G_RAID5_VERSION	3

#ifdef _KERNEL

#define	G_RAID5_TYPE_MANUAL	0
#define	G_RAID5_TYPE_AUTOMATIC	1

#define	G_RAID5_DEBUG(lvl, ...)					\
	do {							\
		if (((int)g_raid5_debug) >= (lvl)) {			\
			printf("GEOM_RAID5");			\
			if ((lvl) > 0) printf("[%u]",lvl);	\
			printf(": ");				\
			printf(__VA_ARGS__);			\
			printf("\n");				\
		}						\
	} while (0)

#define	G_RAID5_LOGREQ(bp, ...)				\
	do {						\
		if (g_raid5_debug >= 5) {		\
			printf("GEOM_RAID5[5]: ");	\
			printf(__VA_ARGS__);		\
			printf(" ");			\
			g_print_bio(bp);		\
			printf("\n");			\
		}					\
	} while (0)

struct g_raid5_softc {
	u_int		sc_type;	/* provider type - automatic or manual */
	struct g_geom	*sc_geom;	/* geom for this device */
#define	sc_name	sc_geom->name
	struct g_provider *sc_provider;	/* provider for this device */
	struct root_hold_token *sc_rootmount;	/* token for root mount hold */
	uint32_t	sc_id;		/* volume unique ID */

	uint16_t	sc_ndisks;	/* total number of disks */
	u_char		lstdno;		/* last disk no? @TODO: need to understand */
	u_char		no_more_open;	/* force close? @TODO: need to understand */
	int		open_count;	/* number of open()s */
	struct mtx	open_mtx;	/* open mutex @TODO: need to understand */
	struct g_consumer **sc_disks;	/* underlying disks */
	char		*sc_disk_states;/* state if disks - ok (0) or error (1) */
#if __FreeBSD_version <= 800001
	struct proc	*workerD;	/* thread for processing finished BIOs */
	struct proc	*worker;	/* main working thread */
#else
	struct proc	*workers_proc;	/* process for this volume */
	struct thread	*workerD;	/* thread for processing finished BIOs */
	struct thread	*worker;	/* main working thread */
#endif
	struct mtx	dq_mtx;		/* mutex for done queue protection */
	struct bio_queue_head dq;	/* done queue - for finished BIOs */
	struct mtx	sq_mtx;		/* mutex for start queue protection */
	struct bio_queue_head sq;	/* start queue - for new BIOs */
	struct bio_queue_head wq;	/* write queue - for write BIOs, awaiting partners */
	struct bio_queue_head wdq;	/* write done queue - for finished writer BIOs */
	
	int		msl;		/* malloc saving queue length */
	int	memuse;		/* used memory */
	struct mtx	msq_mtx;	/* mutex for memory queue protection */
	struct bio_queue_head msq;	/* malloc saving queue */
	
	u_int		wqpi;		/* @TODO: need to understand */
	u_int		wqp;		/* @TODO: need to understand */
	
	struct mtx	sleep_mtx;	/* mutex for sleeping of working thread */
	u_char		no_sleep;	/* @TODO: need to understand */
	u_char		conf_order;	/* reason to write metadata after configuration */
	u_char		no_hot;		/* calm */
	u_char		term;		/* termination phase */
	u_char		hardcoded;	/* is provider hardcoded? */
	u_char		state;		/* state flags */
	u_char		destroy_on_za;	/* some force destroy flag? @TODO: need to understand */
	u_char		stripebits;	/* bits in stripsize */
	uint32_t	stripesize;	/* stripsize as number */
	uint64_t	disksize;	/* size of disk */
	int		veri_pa;	/* verification pause */
	int64_t		verified;	/* number of first potentially dead parity stripe */
	int32_t		newest;		/* number of formerly dead disk */
};
#endif	/* _KERNEL */

#define G_RAID5_STATE_CALM	0x00
#define G_RAID5_STATE_HOT	0x01
#define G_RAID5_STATE_VERIFY	0x02
#define G_RAID5_STATE_SAFEOP	0x04
#define G_RAID5_STATE_COWOP	0x08

struct g_raid5_metadata {
	char		md_magic[16];		/* Magic value. */
	uint32_t	md_version;		/* Version number. */
	char		md_name[16];		/* Volume name. */
	uint32_t	md_id;			/* Unique ID. */
	uint8_t		md_no;			/* Disk number. */
	uint8_t		md_no_hot;		/* no hot disks flag.. */
	uint8_t		md_all;			/* Number of all disks. */
	uint8_t		md_state;		/* Status code */
	uint64_t	md_provsize;		/* Provider's size. */
	uint64_t	md_verified;		/* Verified. */
	uint32_t	md_stripesize;		/* Stripe size. */
	int32_t		md_newest;		/* # of newest disk. */
	char		md_provider[16];	/* Hardcoded provider. */
};

static __inline void
raid5_metadata_encode(const struct g_raid5_metadata *md, u_char *data)
{
	bcopy(md->md_magic, data, sizeof(md->md_magic));
	le32enc(data + 16, md->md_version);
	bcopy(md->md_name, data + 20, sizeof(md->md_name));
	le32enc(data + 36, md->md_id);
	data[40] = md->md_no;
	data[41] = md->md_no_hot;
	data[42] = md->md_all;
	data[43] = md->md_state;
	le64enc(data + 44, md->md_provsize);
	le64enc(data + 52, md->md_verified);
	le32enc(data + 60, md->md_stripesize);
	le32enc(data + 64, md->md_newest);
	bcopy(md->md_provider, data + 68, sizeof(md->md_provider));
}

static __inline void
raid5_metadata_decode(const u_char *data, struct g_raid5_metadata *md)
{
	bcopy(data, md->md_magic, sizeof(md->md_magic));
	md->md_version = le32dec(data + 16);
	bcopy(data + 20, md->md_name, sizeof(md->md_name));
	md->md_id = le32dec(data + 36);
	md->md_no = data[40];
	md->md_no_hot = data[41];
	md->md_all = data[42];
	md->md_state = data[43];
	md->md_provsize = le64dec(data + 44);
	md->md_verified = le64dec(data + 52);
	md->md_stripesize = le32dec(data + 60);
	md->md_newest = le32dec(data + 64);
	bcopy(data + 68, md->md_provider, sizeof(md->md_provider));
}

#endif
